{%- with messages = get_flashed_messages(with_categories=true) -%}
  {% if messages %}
    <ul class="flashes">
    {% for category, message in messages %}
      {% if category == "error" %}
	<li class="{{ category }}">{{ message }}</li>
      {% endif %}
    {% endfor %}
    </ul>
  {% endif %}
{%- endwith %}
